package com.machine.coffee.Activity;

import android.Manifest;
import android.app.Activity;
import android.app.Application;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.app.ProgressDialog;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageManager;
import android.content.DialogInterface;
import android.graphics.PorterDuff;
import android.net.ConnectivityManager;
import android.net.NetworkInfo;
import android.os.Handler;
import android.support.v4.app.ActivityCompat;
import android.support.v4.app.NotificationCompat;
import android.support.v4.content.ContextCompat;
import android.support.v7.app.AlertDialog;
import android.support.v7.app.AppCompatActivity;
import android.os.Bundle;
import android.telephony.TelephonyManager;
import android.util.Log;
import android.view.View;
import android.widget.Button;
import android.widget.CompoundButton;
import android.widget.ImageView;
import android.widget.ProgressBar;
import android.widget.Switch;
import android.widget.TextView;
import android.widget.Toast;

import com.machine.coffee.Controller.AWSConnection;
import com.machine.coffee.R;
import com.machine.coffee.Utils.CoffeeMachinePreferences;
import com.machine.coffee.Utils.MessageTimer;
import com.machine.coffee.Utils.ToastMessage;

import static android.app.PendingIntent.FLAG_UPDATE_CURRENT;

public class MainActivity
        extends AppCompatActivity
        implements CompoundButton.OnCheckedChangeListener {

    // Root topics
    private static final String TOPIC_TURN_ON_OFF = "CoffeeMachine/TurnOnOff";
    private static final String TOPIC_SHORT_COFFEE = "CoffeeMachine/ShortCoffee";
    private static final String TOPIC_LONG_COFFE = "CoffeeMachine/LongCoffee";
    private static final String TOPIC_COFFEE_LEVEL = "CoffeeMachine/CoffeeLevel";
    private static final String TOPIC_WATER_LEVEL = "CoffeeMachine/WaterLevel";
    private static final String TOPIC_GLASS_POSITION = "CoffeeMachine/GlassPosition";
    private static final String TOPIC_ERROR = "CoffeeMachine/Error";

    private static final String IOT = "/IoT";
    private static final String ANDROID = "/Android";

    private static final int READ_PHONE_STATE_PERMISSION = 2;

    //Dialog IDs
    private static final int ERROR_DIALOG_ID = 0;
    private static final int READY_DIALOG_ID = 1;
    private static final int BUSY_DIALOG_ID  = 2;

    //Notification IDs
    private static final int SHORT_COFFEE_ID = 001;
    private static final int LONG_COFFEE_ID = 002;
    private static final int IN_PROGRESS_ID = 003;

    //String which stores the phone's IMEI
    private String mIMEI;

    //Activity status callbacks
    private final Application.ActivityLifecycleCallbacks mActivityCallbacks =
            new ActivityLifecycleCallbackReceiver();

    //Activity status callbacks variables
    private boolean hasActivityPaused, hasActivityStopped, hasActivityResumed;

    //Internet Receiver
    private InternetReceiver mInternetReceiver;

    //Tag to debug
    private static final String LOG_TAG = "DEBUG";

    //Status of water level
    private static final int EMPTY = 0;
    private static final int FULL = 1;

    //Status of Coffee Machine
    private static final String ON = "1";
    private static final String OFF = "0";

    //Status of glass position
    private static final int POSITIONED = 1;
    private static final int NOT_POSITIONED = 0;

    //Turns on/off the Coffee Maker
    private Switch mOnOff;

    //Provides the Activity context
    private Context mContext;

    //String to store the Toast text
    private String mText = "";

    //Dialog and Notification IDs
    private int mShortCoffeeNotifId, mLongCoffeeNotifId, mCoffeeProgressNotifId;

    //NotificationBuilder to create the notifications
    private NotificationCompat.Builder mBuilder;

    //Short and long coffee buttons
    private Button mShortCoffeeButton, mLongCoffeeButton;

    //Progress bar to indicate the coffee level
    private ProgressBar mCoffeeProgressBar;

    //Value of coffee's progress bar
    private int mCoffeeProgressStatus = 0;

    //Boolean to check if the level is enough to make coffee
    private boolean mProgressBarPermission = false;

    //Handler to update the progress bar
    private Handler mCoffeeHandler;

    //TextViews used to indicate the status of each reservoir
    private TextView mCoffeeProgressText, mWaterProgressText, mGlassStatusText;

    //ImageView to show if the cup was positioned
    private ImageView mGlassStatusImage;

    //NotificationManager to manage the notifications
    private NotificationManager mNotifyMgr;

    //Boolean to indicate if the cup was positioned
    private boolean mGlassStatus;

    //Handler used to implement the ProgressDialog timer
    private Handler mHandlerTimer;

    //Values to set the preparation time of each kind of coffee
    private int mShortCoffeeTotalTime = 20;
    private int mLongCoffeeTotalTime = 30;

    //Value to count the time to finish ProgressDialog
    private int mTotalTime;

    //ProgressDialog after clicking on a coffee button
    private ProgressDialog mCoffeeProgressDialog;

    //AWS connection manager object
    private AWSConnection mAWSConnection;

    //Timer class instance
    private MessageTimer mMessageTimer;

    //Boolean to check if Bluetooth Low Energy is connected
    private boolean mBluetoothIsConnected = true;

    //Boolean to check if the coffee requested was finished
    private boolean mCoffeeIsFinished = false;

    //Boolean to check if the coffee is in progress
    private boolean mCoffeeInProgress = false;

    //Boolean to check if the app is being launched for the first time
    private boolean mIsFirstLaunch = true;

    //Boolean to check if the app was already connected before losing connection
    private boolean mLastStateConnected = true;

    //Boolean to check if the permission was granted or not
    private boolean mIsPermissionDenied = false;

    /**
     * Android onCreate method to call actions when the Activity is launched.
     *
     * @param savedInstanceState
     */
    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        initializeCallbacks();
        initializeComponents();
        initInternetReceiver();
        generateDeviceIMEI();
    }

    /**
     * Android onDestroy method called right before the application is closed.
     */
    @Override
    public void onDestroy() {
        super.onDestroy();
        if (mNotifyMgr != null) mNotifyMgr.cancelAll();
        unregisterReceiver(mInternetReceiver);
        getApplication().unregisterActivityLifecycleCallbacks(mActivityCallbacks);
    }

    /**
     * Method to initialize all UI components.
     */
    private void initializeComponents() {
        mContext = getApplicationContext();
        setContentView(R.layout.activity_main);

        startSwitch();
        startNotificationModel();
        startCoffeeButtons();
        startConnectionAWS();
        getTimerInstance();

        mOnOff.setEnabled(false);
        mShortCoffeeButton.setEnabled(false);
        mLongCoffeeButton.setEnabled(false);
    }

    /**
     * Method to start the Internet receiver and check if there is Internet connection.
     */
    private void initInternetReceiver() {
        IntentFilter intentFilter = new IntentFilter();
        intentFilter.addAction("android.net.conn.CONNECTIVITY_CHANGE");
        mInternetReceiver = new InternetReceiver();
        registerReceiver(mInternetReceiver, intentFilter);
    }

    /**
     * Method to get all ActivityLifecycleCallbacks.
     */
    public void initializeCallbacks() {
        getApplication().registerActivityLifecycleCallbacks(mActivityCallbacks);
    }

    /**
     * Method to connect with AWS if there is Internet connection.
     */
    private void startConnectionAWS() {
        mAWSConnection = AWSConnection.getInstance(this);
        if (hasInternetConnection()) {
            mAWSConnection.getConnection();
        }
        mAWSConnection.setLastStatus();
    }

    /**
     * Method to get MessageTimer instance using Singleton pattern.
     */
    private void getTimerInstance() {
        mMessageTimer = MessageTimer.getInstance(this);
    }

    /**
     * Method to set the dialog according to the status of the coffee. This method is called when
     * a coffee is requested or finished.
     *
     * @param status is a variable which indicates if a coffee is ready or not and also notifies if
     *               an error occurs.
     */
    private void coffeeReadyDialog(int status) {
        AlertDialog.Builder readyDialog = new AlertDialog.Builder(MainActivity.this);

        if (status == ERROR_DIALOG_ID) { //error
            readyDialog.setTitle(R.string.ops).setMessage(R.string.connection_error);
        } else if (status == READY_DIALOG_ID) { //coffee ready
            readyDialog.setTitle(R.string.coffee_ready).setMessage(R.string.coffee_ready_message);
        } else if (status == BUSY_DIALOG_ID) { //coffee in progress
            readyDialog.setTitle(getString(R.string.app_name)).
                    setMessage(getString(R.string.machine_in_use));
        }

        readyDialog.setCancelable(false)
                .setPositiveButton(R.string.ok, new DialogInterface.OnClickListener() {
                    @Override
                    public void onClick(DialogInterface dialogInterface, int i) {
                        //When OK is pressed, all notifications are closed
                        if (mNotifyMgr != null) mNotifyMgr.cancelAll();
                    }
                });
        AlertDialog dialog = readyDialog.create();
        dialog.show();
    }

    /**
     * Method to set the switch according to the Coffee Maker status. If the Coffee Maker is turned
     * on by another phone, the status is updated on the whole connected devices.
     *
     * @param status is a String which tells if the Coffee Maker is on or off.
     */
    public void setCheckSwitch(String status) {
        mOnOff.setClickable(true);
        if ((status.equals(ON))) {
            if ((CoffeeMachinePreferences.getTurnOnOff(getApplicationContext()).equals(OFF))) {
                mOnOff.setOnCheckedChangeListener(null);
                mOnOff.setChecked(true);
                mOnOff.setOnCheckedChangeListener(this);
                ToastMessage.setToastMessage(mContext,
                        getString(R.string.coffee_machine_on),
                        Toast.LENGTH_LONG);
            } else if (mIsFirstLaunch && (status.equals(ON)) &&
                    (!mOnOff.isChecked())) {
                mOnOff.setOnCheckedChangeListener(null);
                mOnOff.setChecked(true);
                mOnOff.setOnCheckedChangeListener(this);
                ToastMessage.setToastMessage(mContext,
                        getString(R.string.coffee_machine_on),
                        Toast.LENGTH_LONG);
            }
        } else if ((status.equals(OFF))) {
            if ((CoffeeMachinePreferences.getTurnOnOff(getApplicationContext()).equals(ON))) {
                mOnOff.setOnCheckedChangeListener(null);
                mOnOff.setChecked(false);
                mOnOff.setOnCheckedChangeListener(this);
                ToastMessage.setToastMessage(mContext,
                        getString(R.string.coffee_machine_off),
                        Toast.LENGTH_LONG);
            } else if (mIsFirstLaunch) {
                mOnOff.setOnCheckedChangeListener(null);
                mOnOff.setChecked(false);
                mOnOff.setOnCheckedChangeListener(this);
                ToastMessage.setToastMessage(mContext,
                        getString(R.string.coffee_machine_off),
                        Toast.LENGTH_LONG);
            }
        } else if (status.equals("busy")) {
            mOnOff.setChecked(true);
        }
        if (mIsFirstLaunch && !mIsPermissionDenied) {
            enableComponents();
            mIsFirstLaunch = false;
        }
    }

    /**
     * Method which enables all the components to allow the user to use the Coffee Maker.
     */
    public void enableComponents() {
        mOnOff.setEnabled(true);
        mShortCoffeeButton.setEnabled(true);
        mLongCoffeeButton.setEnabled(true);
    }

    /**
     * Method which disables all the components to do not allow the user to use the Coffee Maker;
     */
    public void disableComponents() {
        mOnOff.setEnabled(false);
        mShortCoffeeButton.setEnabled(false);
        mLongCoffeeButton.setEnabled(false);
    }

    /**
     * Method to instantiate the Switch which turns on/off the Coffee Maker.
     */
    private void startSwitch() {
        mOnOff = (Switch) findViewById(R.id.on_off_switch);
        mOnOff.setClickable(false);
        mOnOff.setOnCheckedChangeListener(this);
    }

    /**
     * Method to instantiate and set the OnClickListeners for each coffee button. When a button is
     * pressed, a message is sent on TOPIC_SHORT_COFFEE or TOPIC_LONG_COFFEE to notify the other
     * users that coffee is being made and to wait for an answer from the DragonBoard.
     */
    private void startCoffeeButtons() {
        mShortCoffeeButton = (Button) findViewById(R.id.short_coffee_button);
        mLongCoffeeButton = (Button) findViewById(R.id.long_coffee_button);

        mShortCoffeeButton.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {

                if (isEverythingOK()) {
                    mOnOff.setEnabled(false);
                    mLongCoffeeButton.setEnabled(false);

                    coffeeProgressDialog(mShortCoffeeNotifId);

                    mAWSConnection.topicPublish(mIMEI, TOPIC_SHORT_COFFEE + ANDROID);
                }
            }
        });

        mLongCoffeeButton.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View view) {
                if (isEverythingOK()) {
                    mOnOff.setEnabled(false);
                    mShortCoffeeButton.setEnabled(false);

                    coffeeProgressDialog(mLongCoffeeNotifId);
                    mAWSConnection.topicPublish(mIMEI, TOPIC_LONG_COFFE + ANDROID);
                }
            }
        });
    }

    /**
     * Method to instantiate and update the ProgressBar which displays the amount of coffee
     * available. The app will only accept coffee requests if mCoffeeProgressStatus is above 5%.
     *
     * @param data variable where the message containing the progress value is stored.
     */
    public void startCoffeeProgressBar(final int data) {
        mCoffeeProgressBar = (ProgressBar) findViewById(R.id.coffee_progress_bar);
        mCoffeeHandler = new Handler();
        mCoffeeProgressText = (TextView) findViewById(R.id.coffee_status_text);

        new Thread(new Runnable() {
            public void run() {
                mProgressBarPermission = true;

                mCoffeeProgressStatus = data;

                runOnUiThread(new Runnable() {
                    @Override
                    public void run() {

                        if (mCoffeeProgressStatus <= 100 && mCoffeeProgressStatus >= 50) {
                            mCoffeeProgressBar.getProgressDrawable().
                                    setColorFilter(ContextCompat.getColor(getApplicationContext()
                                            , R.color.progressGreen),
                                            PorterDuff.Mode.SRC_IN);
                        } else if (mCoffeeProgressStatus < 50 && mCoffeeProgressStatus >= 20) {
                            mCoffeeProgressBar.getProgressDrawable().
                                    setColorFilter(ContextCompat.getColor(getApplicationContext()
                                            , R.color.progressYellow),
                                            PorterDuff.Mode.SRC_IN);
                        } else if (mCoffeeProgressStatus < 20 && mCoffeeProgressStatus >= 5) {
                            mCoffeeProgressBar.getProgressDrawable().
                                    setColorFilter(ContextCompat.getColor(getApplicationContext()
                                            , R.color.progressRed),
                                            PorterDuff.Mode.SRC_IN);
                        } else {
                            mCoffeeProgressBar.getProgressDrawable().
                                    setColorFilter(ContextCompat.getColor(getApplicationContext()
                                            , R.color.progressRed),
                                            PorterDuff.Mode.SRC_IN);
                            mProgressBarPermission = false;
                        }

                        mCoffeeProgressText.setText(mCoffeeProgressStatus + "%");
                    }
                });

                //Updates the progress bar
                mCoffeeHandler.post(new Runnable() {
                    @Override
                    public void run() {
                        mCoffeeProgressBar.setProgress(mCoffeeProgressStatus);
                    }
                });
            }
        }).start();
    }

    /**
     * Method to set the TextView which notifies the user about the status of the water reservoir.
     *
     * @param waterProgressStatus is the variable which contains the status of the water reservoir.
     *                            It can be FULL, or EMPTY, according to what is sent by the
     *                            Dragonboard.
     */
    public void startWaterStatus(int waterProgressStatus) {
        mWaterProgressText = (TextView) findViewById(R.id.water_status_text);

        if (waterProgressStatus == FULL) {
            mWaterProgressText.setText(R.string.full);
            mWaterProgressText.setTextColor(ContextCompat.getColor(this, R.color.progressGreen));
        } else if (waterProgressStatus == EMPTY) {
            mWaterProgressText.setText(R.string.empty);
            mWaterProgressText.setTextColor(ContextCompat.getColor(this, R.color.progressRed));
        }
    }

    /**
     * Method to create the NotificationCompat.Builder object and set a notification model to use
     * on the app notifications.
     */
    private void startNotificationModel() {
        mShortCoffeeNotifId = SHORT_COFFEE_ID;
        mLongCoffeeNotifId = LONG_COFFEE_ID;
        mCoffeeProgressNotifId = IN_PROGRESS_ID;

        mBuilder = new NotificationCompat.Builder(MainActivity.this)
                .setSmallIcon(R.mipmap.icon_launcher)
                .setContentTitle(getString(R.string.app_name))
                .setAutoCancel(true);

        //Pending Intent is used to define the action that will occur after clicking on notification
        Intent resultIntent = new Intent(this, MainActivity.class);

        PendingIntent resultPendingIntent =
                PendingIntent.getActivity(
                        this,
                        0,
                        resultIntent,
                        FLAG_UPDATE_CURRENT
                );

        mBuilder.setContentIntent(resultPendingIntent);
    }

    /**
     * Method to send a notification according to its ID. The variables mCoffeeInProgress and
     * mCoffeeIsFinished are set as false to indicate that the progress dialog was dismissed and
     * the process can be started again.
     *
     * @param mNotifId is a variable which contains the notification ID. The values can be:
     *                 - mShortCoffeeNotifId    = SHORT_COFFEE_ID
     *                 - mLongCoffeeNotifId     = LONG_COFFEE_ID
     *                 - mCoffeeProgressNotifId = IN_PROGRESS_ID
     */
    private void sendNotification(int mNotifId) {
        mCoffeeInProgress = false;
        mCoffeeIsFinished = false;

        mNotifyMgr = (NotificationManager) getSystemService(NOTIFICATION_SERVICE);
        mNotifyMgr.notify(mNotifId, mBuilder.build());
    }

    /**
     * Method to instantiate and set the UI elements related to the cup status indication.
     *
     * @param status is a variable to notify the cup status. It can be POSITIONED or NOT_POSITIONED.
     */
    public void startGlassStatus(int status) {
        mGlassStatusText = (TextView) findViewById(R.id.glass_indicator_text);
        mGlassStatusImage = (ImageView) findViewById(R.id.glass_indicator_image);

        if (status == POSITIONED) {
            mGlassStatus = true;
        } else if (status == NOT_POSITIONED) {
            mGlassStatus = false;
        }

        if (mGlassStatus) {
            //If the cup is positioned
            mGlassStatusText.setText(R.string.yes_glass);
            mGlassStatusText.setTextColor(ContextCompat.getColor(this, R.color.progressGreen));

            mGlassStatusImage.setAlpha(1.0f);
        } else {
            //If the cup is not positioned
            mGlassStatusText.setText(R.string.no_glass);
            mGlassStatusText.setTextColor(ContextCompat.getColor(this, R.color.progressRed));

            mGlassStatusImage.setAlpha(0.1f);
        }
    }

    /**
     * Method to instantiate and set the dialog which indicates that the coffee is being prepared.
     * Then, the Runnable class is called to start the timer for each kind of coffee and when the
     * coffee is finished, the notification and the dialog are set according to what was asked to
     * the user.
     *
     * @param id is a variable to indicate what kind of information will be handled on the dialog
     *           and the notification.
     */
    private void coffeeProgressDialog(int id) {
        final int coffeeId = id;

        mCoffeeProgressDialog = new ProgressDialog(this);
        mCoffeeProgressDialog.setTitle(R.string.app_name);

        switch (coffeeId) {
            case SHORT_COFFEE_ID:
                mTotalTime = mShortCoffeeTotalTime;
                mCoffeeProgressDialog.setMessage(getString(R.string.making_short_coffee));
                break;
            case LONG_COFFEE_ID:
                mTotalTime = mLongCoffeeTotalTime;
                mCoffeeProgressDialog.setMessage(getString(R.string.making_long_coffee));
                break;
            default:
                break;
        }

        mCoffeeProgressDialog.setProgressStyle(ProgressDialog.STYLE_SPINNER);
        mCoffeeProgressDialog.setCancelable(false);
        mCoffeeProgressDialog.show();

        mBuilder.setContentText(getString(R.string.making_coffee));
        mBuilder.setOngoing(false);

        sendNotification(mCoffeeProgressNotifId);

        mHandlerTimer = new Handler();
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                try {
                    mTotalTime--;

                    if (wasCoffeeFinished()) {
                        //Coffee ready
                        mCoffeeProgressDialog.dismiss();
                        coffeeReadyDialog(1); //Notifies that the coffee is ready

                        if (mNotifyMgr != null) mNotifyMgr.cancelAll();

                        if (coffeeId == SHORT_COFFEE_ID) mBuilder.
                                setContentText(getString(R.string.short_coffee_ready));
                        else if (coffeeId == LONG_COFFEE_ID) mBuilder.
                                setContentText(getString(R.string.long_coffee_ready));

                        mOnOff.setEnabled(true);
                        mShortCoffeeButton.setEnabled(true);
                        mLongCoffeeButton.setEnabled(true);

                        sendNotification(coffeeId);
                    } else if (mTotalTime <= 0) {
                        //Time limit exceeded
                        mCoffeeProgressDialog.dismiss();
                        coffeeReadyDialog(0); //Notifies the error

                        if (mNotifyMgr != null) mNotifyMgr.cancelAll();

                        mBuilder.setContentText(getString(R.string.error_communication));

                        mOnOff.setEnabled(true);
                        mShortCoffeeButton.setEnabled(true);
                        mLongCoffeeButton.setEnabled(true);

                        sendNotification(coffeeId);
                    } else if (isCoffeeInProgress()) {
                        //Coffee machine in use
                        mCoffeeProgressDialog.dismiss();
                        coffeeReadyDialog(2); //Notifies that the coffee machine is being used

                        if (mNotifyMgr != null) mNotifyMgr.cancelAll();

                        mBuilder.setContentText(getString(R.string.machine_in_use));

                        mOnOff.setEnabled(true);
                        mShortCoffeeButton.setEnabled(true);
                        mLongCoffeeButton.setEnabled(true);

                        sendNotification(coffeeId);
                    } else {
                        //Timer error
                        mHandlerTimer.postDelayed(this, 1000);
                    }
                }
                catch (Exception e) {
                    //The Runnable thread was interrupted.
                }
            }
        };
        mHandlerTimer.postDelayed(runnable, 1000);
    }

    /**
     * Method to turn off the switch.
     */
    public void turnOffSwitch() {
        mOnOff.setChecked(false);
        mOnOff.setEnabled(true);
    }

    /**
     * Method to cancel the TimerTask and enable the switch.
     */
    public void coffeeMachineState() {
        mMessageTimer.cancelTimerTask();
        mOnOff.setEnabled(true);
    }

    /**
     * Before making coffee, the system checks if everything is ok. This verification is divided on
     * two levels.
     *
     * @return
     */
    public boolean isEverythingOK() {
        //First level
        if (systemIsReady()) {
            //Second level
            if (coffeeMakerIsReady()) {
                return true;
            } else {
              //The coffee machine is not ready to prepare coffee
            }
        } else {
            //Communication or connection error
            ToastMessage.setToastMessage(mContext,
                    getString(R.string.connection_error), Toast.LENGTH_SHORT);

        }
        return false;
    }

    /**
     * The first level of verification before making coffee. It checks:
     * - If there is Internet connection
     * - If there is AWS connection
     * - If there is Bluetooth connection
     *
     * @return is the status of the first level of verification
     */
    public boolean systemIsReady() {
        if (hasInternetConnection()) {
          //Internet is OK
            if (mAWSConnection.isConnected()) {
                //AWS connection is OK
                if (hasBluetoothConnection()) {
                    //Bluetooth is OK
                    return true;
                } else {
                    //Bluetooth disconnected
                }
            } else {
                //AWS disconnected
            }
        } else {
            //Internet disconnected
        }
        return false;
    }

    /**
     * The second level of verification before making coffee. It checks:
     * - If coffee machine was turned on
     * - If the coffee machine is not being used by another user
     * - If the cup is positioned
     * - If there is enough coffee
     * - If there is enough water
     *
     * @return is the status of the second level of verification
     */
    public boolean coffeeMakerIsReady() {
        if (mOnOff.isChecked()) {
          //The coffee machine was turned on
            if (!isCoffeeInProgress()) {
                //The coffee machine is available to use
                if (mAWSConnection.isGlassPositioned()) {
                    //The glass is positioned
                    if (isCoffeeLevelEnough()) {
                        //The coffee level is ok
                        if (!mAWSConnection.isWaterLevelEmpty()) {
                            //The water level is enough
                            return true;
                        } else {
                            //The water level is not enough
                            ToastMessage.setToastMessage(mContext,
                                    getString(R.string.water_level_empty), Toast.LENGTH_SHORT);
                        }
                    } else {
                        //The coffee level is not enough
                        ToastMessage.setToastMessage(mContext,
                                getString(R.string.coffee_level_empty), Toast.LENGTH_SHORT);
                    }
                } else {
                    //The glass is not positioned
                    ToastMessage.setToastMessage(mContext,
                            getString(R.string.no_glass), Toast.LENGTH_SHORT);
                }
            } else {
                //The coffee machine is in use
                ToastMessage.setToastMessage(mContext,
                        getString(R.string.doing_coffee), Toast.LENGTH_LONG);
            }
        } else {
            //The coffee machine was turned off
            ToastMessage.setToastMessage(mContext,
                    getString(R.string.coffee_machine_off), Toast.LENGTH_LONG);
        }
        return false;
    }

    /**
     * Method to check if there is Internet connection. If there is not, the app disconnects from
     * AWS. If then the phone reconnects to a valid network, is reconnects to AWS.
     *
     * @return if the device is connected to a valid network.
     */
    public boolean hasInternetConnection() {
        ConnectivityManager cm =
                (ConnectivityManager) mContext.
                        getSystemService(Context.CONNECTIVITY_SERVICE);

        NetworkInfo activeNetwork = cm.getActiveNetworkInfo();
        boolean isConnected = activeNetwork != null &&
                activeNetwork.isConnectedOrConnecting();

        if (!isConnected) {
            ToastMessage.setToastMessage(mContext,
                    getString(R.string.no_internet) + " " + getString(R.string.please_connect),
                    Toast.LENGTH_LONG);
            mLastStateConnected = false;
            mAWSConnection.disconnectAWS();
        } else {
            if (!mLastStateConnected) {
                mAWSConnection.getConnection();
                ToastMessage.setToastMessage(
                        mContext,
                        getString(R.string.internet_connected),
                        Toast.LENGTH_SHORT);
            }
            mLastStateConnected = true;
        }

        return isConnected;
    }

    /**
     * Method to set the variable which notifies if there is Bluetooth connection. The DragonBoard
     * sends a message through TOPIC_ERROR if it cannot connects to CSR1011. The two acceptable
     * messages to assume that there is no Bluetooth connection are "write_char_error" and
     * "connection_error".
     *
     * @param message is a variable which contains the message sent by the DragonBoard.
     */
    public void checkBluetoothConnection(String message) {
        boolean isBluetoothConnected;

        if (message.equals("write_char_error") || message.equals("connection_error")) {
            isBluetoothConnected = false;
            disableComponents();
        } else {
            isBluetoothConnected = true;
            enableComponents();
        }

        this.mBluetoothIsConnected = isBluetoothConnected;
    }

    /**
     * Method to check (inside the application) the Bluetooth connection.
     *
     * @return the status of the Bluetooth connection.
     */
    public boolean hasBluetoothConnection() {
        return mBluetoothIsConnected;
    }

    /**
     * Method to set the variables which indicate the status of the machine (isCoffeeFinished and
     * isCoffeeInProgress). The IMEI is used here to notify the user if the coffee machine is busy.
     *
     * @param message is a variable which stores the message received when this method is called
     */
    public void checkIfCoffeeIsFinished(String message) {
        boolean isCoffeeFinished = false, isCoffeeInProgress = false;

        if (message.equals("coffee_ready" + mIMEI)) {
            isCoffeeFinished = true;
            isCoffeeInProgress = false;
        } else if (message.equals(TOPIC_ERROR + IOT)) {
            isCoffeeFinished = false;
            isCoffeeInProgress = false;
        } else if (message.equals(mIMEI)) {
            //if the IMEI received is the same retrieved from this device, the machine is busy
            isCoffeeInProgress = true;
        } else if (message.equals(TOPIC_TURN_ON_OFF + ANDROID)) {
            ToastMessage.setToastMessage(this, getString(R.string.coffee_maker_error),
                    Toast.LENGTH_SHORT);
        } else {
            //IMEI is not the same
        }

        this.mCoffeeIsFinished = isCoffeeFinished;
        this.mCoffeeInProgress = isCoffeeInProgress;
    }

    /**
     * Method to check if the coffee was finished.
     *
     * @return if the coffee was finished.
     */
    public boolean wasCoffeeFinished() {
        return mCoffeeIsFinished;
    }

    /**
     * Method to check if the coffee is in progress.
     *
     * @return if the coffee is in progress.
     */
    public boolean isCoffeeInProgress() {
        return mCoffeeInProgress;
    }

    /**
     * Method to check if the coffee level is enough to make coffee.
     *
     * @return if the level is enough to make coffee.
     */
    public boolean isCoffeeLevelEnough() {
        if (mAWSConnection.isCoffeeLevelFull() && mProgressBarPermission) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Method which is called when the status of the switch changes. To avoid the status of the
     * hardware to be different from the app, some verifications are made before switching it on or
     * off.
     *
     * @param buttonView contains the XML element.
     * @param isChecked contais the status of the switch.
     */
    @Override
    public void onCheckedChanged(CompoundButton buttonView, boolean isChecked) {
        if (hasInternetConnection()) {
            if (isChecked) {
                if (mAWSConnection.isConnected() && hasBluetoothConnection()) {
                    mOnOff.setEnabled(false);
                    mMessageTimer.startTimerTask();
                    mAWSConnection.topicPublish(String.valueOf(ON),
                            TOPIC_TURN_ON_OFF + ANDROID);
                } else {
                    mText = getString(R.string.lost_connection);
                    mOnOff.setChecked(false);
                }
            } else {
                if (mAWSConnection.isConnected() && hasBluetoothConnection()) {
                    mAWSConnection.topicPublish(String.valueOf(OFF),
                            TOPIC_TURN_ON_OFF + ANDROID);
                } else {
                    mText = getString(R.string.lost_connection);
                    mOnOff.setChecked(true);
                }
            }
        } else {
            if (isChecked) {
                mText = getString(R.string.lost_connection);
                mOnOff.setChecked(false);
            } else {
                mText = getString(R.string.lost_connection);
                mOnOff.setChecked(true);
            }
        }
    }

    /**
     * Method to ask permission to retrieve the device IMEI.
     */
    public void generateDeviceIMEI() {
        if (ContextCompat.checkSelfPermission(this,
                Manifest.permission.READ_PHONE_STATE)
                != PackageManager.PERMISSION_GRANTED) {
            //Permission denied
            ActivityCompat.requestPermissions(this,
                    new String[]{Manifest.permission.READ_PHONE_STATE},
                    READ_PHONE_STATE_PERMISSION);
        } else {
            getDeviceId();
        }
    }

    /**
     * Method to receive the results of permission requests.
     *
     * @param requestCode is passed on requestPermissions(android.app.Activity, String[], int).
     * @param permissions retrieves the requested permissions.
     * @param grantResults retrieves the results, which can be PERMISSION_GRANTED or
     *                     PERMISSION_DENIED.
     */
    @Override
    public void onRequestPermissionsResult(int requestCode,
                                           String permissions[],
                                           int[] grantResults) {
        switch (requestCode) {
            case READ_PHONE_STATE_PERMISSION: {
                // If request is cancelled, the result arrays are empty.
                if (grantResults.length > 0
                        && grantResults[0] == PackageManager.PERMISSION_GRANTED) {
                    getDeviceId();
                    mIsPermissionDenied = false;
                } else {
                    // PERMISSÃO NEGADA
                    mIsPermissionDenied = true;
                    mIsFirstLaunch = true;
                    AlertDialog.Builder readyDialog = new AlertDialog.Builder(MainActivity.this);
                    readyDialog.setTitle(getString(R.string.warning))
                            .setMessage(getString(R.string.app_require_permission))
                            .setPositiveButton(R.string.ok, new DialogInterface.OnClickListener() {
                                @Override
                                public void onClick(DialogInterface dialog, int which) {

                                }
                            });
                    readyDialog.show();
                }
                return;
            }
        }
    }

    /**
     * Method to generate the device IMEI to notify the other users if the machine is busy.
     */
    public void getDeviceId() {
        TelephonyManager manager = (TelephonyManager)
                mContext.getSystemService(Context.TELEPHONY_SERVICE);

        this.mIMEI = manager.getDeviceId();
        Log.d(LOG_TAG, this.mIMEI);
    }

    /**
     * Method to generate the activity's lifecycle callbacks. When the activity follows its
     * lifecycle, each state is covered here to facilitate the implementation of some actions that
     * must occur on each of them.
     */
    public class ActivityLifecycleCallbackReceiver implements
            Application.ActivityLifecycleCallbacks {
        @Override
        public void onActivityCreated(Activity activity, Bundle savedInstanceState) {

        }

        @Override
        public void onActivityStarted(Activity activity) {

        }

        @Override
        public void onActivityResumed(Activity activity) {
            hasActivityResumed = true;
            hasActivityPaused = false;
            if (!mOnOff.isEnabled() && !mShortCoffeeButton.isEnabled()
                    && !mLongCoffeeButton.isEnabled() && !mIsFirstLaunch) {
                enableComponents();
                mIsFirstLaunch = true;
                mIsPermissionDenied = false;
            }
        }

        @Override
        public void onActivityPaused(Activity activity) {
            hasActivityPaused = true;
            hasActivityResumed = false;
        }

        @Override
        public void onActivitySaveInstanceState(Activity activity, Bundle outState) {

        }

        @Override
        public void onActivityStopped(Activity activity) {
            hasActivityStopped = true;
            hasActivityResumed = false;
            mIsFirstLaunch = false;
        }

        @Override
        public void onActivityDestroyed(Activity activity) {

        }
    }

    /**
     *  Class to implement the InternetReceiver and check if there is InternetConnection.
     */
    public class InternetReceiver extends BroadcastReceiver {
        @Override
        public void onReceive(Context context, Intent intent) {
            hasInternetConnection();
        }
    }
}
